import sys
from dataclasses import dataclass
from typing import Callable, List, Union, Dict, Optional
import dataclasses
import numpy as np
import torch
import onnx
from turnkeyml.common import printing
import turnkeyml.common.build as build
from turnkeyml.common.performance import MeasuredPerformance
import turnkeyml.common.filesystem as fs


class AnalysisException(Exception):
    """
    Indicates a failure during analysis
    """


@dataclass
class UniqueInvocationInfo:
    """
    Refers to unique static model invocations
    (i.e. models executed with unique input shapes)
    """

    hash: Union[str, None] = None
    parent_hash: Union[str, None] = None
    performance: MeasuredPerformance = None
    traceback: List[str] = None
    inputs: Union[dict, None] = None
    input_shapes: Union[dict, None] = None
    executed: int = 0
    exec_time: float = 0.0
    status_message: str = ""
    is_target: bool = False
    status_message_color: printing.Colors = printing.Colors.ENDC
    traceback_message_color: printing.Colors = printing.Colors.FAIL
    stats_keys: Optional[List[str]] = None
    stats: fs.Stats = None


@dataclass
class ModelInfo:
    model: torch.nn.Module
    name: str
    script_name: str
    file: str = ""
    line: int = 0
    params: int = 0
    depth: int = 0
    hash: Union[str, None] = None
    parent_hash: Union[str, None] = None
    old_forward: Union[Callable, None] = None
    unique_invocations: Union[
        Dict[str, UniqueInvocationInfo], None
    ] = dataclasses.field(default_factory=dict)
    last_unique_invocation_executed: Union[str, None] = None
    build_model: bool = False
    model_type: build.ModelType = build.ModelType.PYTORCH

    def __post_init__(self):
        self.params = count_parameters(self.model, self.model_type)


def count_parameters(model: torch.nn.Module, model_type: build.ModelType) -> int:
    """
    Returns the number of parameters of a given model
    """
    if model_type == build.ModelType.PYTORCH:
        return sum([parameter.numel() for _, parameter in model.named_parameters()])
    elif model_type == build.ModelType.KERAS:
        return model.count_params()
    elif model_type == build.ModelType.ONNX_FILE:
        onnx_model = onnx.load(model)
        return int(
            sum(
                np.prod(tensor.dims)
                for tensor in onnx_model.graph.initializer
                if tensor.name not in onnx_model.graph.input
            )
        )

    # Raise exception if an unsupported model type is provided
    raise AnalysisException(f"model_type {model_type} is not supported")


def get_onnx_ops_list(onnx_model) -> Dict:
    """
    List unique ops found in the onnx model
    """
    onnx_ops_counter = {}
    try:
        model = onnx.load(onnx_model)
    except Exception as e:  # pylint: disable=broad-except
        printing.log_warning(f"Failed to get ONNX ops list from {onnx_model}: {str(e)}")
        return onnx_ops_counter
    for node in model.graph.node:  # pylint: disable=E1101
        onnx_ops_counter[node.op_type] = onnx_ops_counter.get(node.op_type, 0) + 1
    return onnx_ops_counter


def populate_onnx_model_info(onnx_model) -> Dict:
    """
    Read the model metadata to populate IR, Opset and model size
    """
    result_dict = {
        "ir_version": None,
        "opset": None,
        "size on disk (KiB)": None,
    }
    try:
        model = onnx.load(onnx_model)
    except Exception as e:  # pylint: disable=broad-except
        printing.log_warning(f"Failed to get ONNX ops list from {onnx_model}: {str(e)}")
        result_dict.update({"error": "ONNX model analysis failed"})
        return result_dict
    # pylint: disable=E1101
    result_dict.update(
        {
            "ir_version": getattr(model, "ir_version", None),
            "opset": getattr(model.opset_import[0], "version", None),
        }
    )
    try:
        result_dict.update(
            {
                "size on disk (KiB)": round(
                    model.SerializeToString().__sizeof__() / 1024, 4
                ),
            }
        )
    except ValueError:
        # Models >2GB on disk cannot have their model size measured this
        # way and will throw a ValueError https://github.com/onnx/turnkeyml/issues/41
        pass

    return result_dict


def onnx_input_dimensions(onnx_model) -> Dict:
    """
    Read model input dimensions
    """
    input_shape = {}
    try:
        model = onnx.load(onnx_model)
    except Exception as e:  # pylint: disable=broad-except
        printing.log_warning(f"Failed to get ONNX ops list from {onnx_model}: {str(e)}")
        return input_shape
    for inp in model.graph.input:  # pylint: disable=E1101
        shape = str(inp.type.tensor_type.shape.dim)
        input_shape[inp.name] = [int(s) for s in shape.split() if s.isdigit()]
    return input_shape


def stop_logger_forward() -> None:
    """
    Stop forwarding stdout and stderr to file
    """
    if hasattr(sys.stdout, "terminal"):
        sys.stdout = sys.stdout.terminal
    if hasattr(sys.stderr, "terminal_err"):
        sys.stderr = sys.stderr.terminal_err


def analyze_onnx(build_name: str, cache_dir: str, stats: fs.Stats):
    # ONNX stats that we want to save into the build's turnkey_stats.yaml file
    # so that they can be easily accessed by the report command later
    if fs.Keys.ONNX_FILE in stats.evaluation_stats.keys():
        # Just in case the ONNX file was generated on a different machine:
        # strip the state's cache dir, then prepend the current cache dir
        final_onnx_file = fs.rebase_cache_dir(
            stats.evaluation_stats[fs.Keys.ONNX_FILE],
            build_name,
            cache_dir,
        )

        onnx_ops_counter = get_onnx_ops_list(final_onnx_file)
        onnx_model_info = populate_onnx_model_info(final_onnx_file)
        input_dimensions = onnx_input_dimensions(final_onnx_file)

        stats.save_model_stat(
            fs.Keys.ONNX_OPS_COUNTER,
            onnx_ops_counter,
        )
        stats.save_model_stat(
            fs.Keys.ONNX_MODEL_INFO,
            onnx_model_info,
        )
        stats.save_model_stat(
            fs.Keys.ONNX_INPUT_DIMENSIONS,
            input_dimensions,
        )



import turnkeyml.build.sequences as sequences
from .runtime import OnnxRT

implements = {
    "runtimes": {
        "ort": {
            "build_required": True,
            "RuntimeClass": OnnxRT,
            "supported_devices": {"x86"},
            "default_sequence": sequences.optimize_fp32,
        }
    }
}



# labels: test_group::turnkey name::chebconv author::graph_convolutions task::Graph_Machine_Learning license::mit
"""
The chebyshev spectral graph convolutional operator from the `"Convolutional
Neural Networks on Graphs with Fast Localized Spectral Filtering"
<https://arxiv.org/abs/1606.09375>`_ paper
"""
from turnkeyml.parser import parse
import torch

from torch_geometric.datasets import Planetoid
from torch_geometric.nn import ChebConv

dataset = Planetoid(root=".", name="Cora")
data = dataset[0]
edge_index_rows = 2

# Parsing command-line arguments
k = parse(["k"])[0]

model = ChebConv(
    dataset.num_features, dataset.num_classes, k
)  # k - Chebyshev filter size
x = torch.ones(data.num_nodes, data.num_features, dtype=torch.float)
edge_index = torch.ones(edge_index_rows, data.num_nodes, dtype=torch.long)
inputs = {
    "x": x,
    "edge_index": edge_index,
}

# Call model
model(**inputs)



from typing import Dict, List, Optional
import turnkeyml.common.exceptions as exp


def decode_args(args: Optional[List[str]]) -> Dict:
    """
    Decode CLI arguments that contain the format
    ["arg1::[value1,value2]", "arg2::value1", "flag_arg"]
    """
    arg_dict = dict()
    if args is None:
        return arg_dict
    for arg in args:
        split_arg = arg.split("::")
        if len(split_arg) == 1:
            # Arguments without values (flags)
            arg_dict[split_arg[0]] = True
        elif len(split_arg) == 2:
            if split_arg[1].startswith("[") and split_arg[1].endswith("]"):
                # Arguments with lists of values
                arg_dict[split_arg[0]] = split_arg[1][1:-1].split(",")
            else:
                # Arguments with single value
                arg_dict[split_arg[0]] = split_arg[1]
        else:
            raise exp.ArgError(
                (
                    f"Coudn't decode rt-arg argument {arg}. The full set "
                    f"of args received was: {args}. Please make sure "
                    "that your rt-arg arguments adhere to the format "
                    "'arg1::value1,value2 arg2::value1 flag_arg'"
                )
            )
    return arg_dict


def encode_args(args: Dict[str, Optional[List]]) -> List[str]:
    """
    Encode Dict into CLI arguments in the format
    ["arg1::[value1,value2]", "arg2::value1", "flag_arg"]
    """
    if args:
        encoded_dict = []
        for key, value in args.items():
            if value is True:
                encoded_dict.append(key)
            elif not isinstance(value, list):
                encoded_dict.append(f"{key}::{value}")
            else:
                encoded_dict.append(f"{key}::[{','.join(value)}]")
        return encoded_dict
    else:
        return []



import platform
import os
import numpy as np
from turnkeyml.run.basert import BaseRT
import turnkeyml.common.exceptions as exp
from turnkeyml.run.onnxrt.execute import ORT_VERSION
from turnkeyml.common.filesystem import Stats
from turnkeyml.run.onnxrt.execute import create_conda_env, execute_benchmark
import turnkeyml.run.plugin_helpers as plugin_helpers


class OnnxRT(BaseRT):
    def __init__(
        self,
        cache_dir: str,
        build_name: str,
        stats: Stats,
        iterations: int,
        device_type: str,
        runtime: str = "ort",
        tensor_type=np.array,
        model=None,
        inputs=None,
    ):
        super().__init__(
            cache_dir=cache_dir,
            build_name=build_name,
            stats=stats,
            tensor_type=tensor_type,
            device_type=device_type,
            iterations=iterations,
            runtime=runtime,
            runtimes_supported=["ort"],
            runtime_version=ORT_VERSION,
            base_path=os.path.dirname(__file__),
            model=model,
            inputs=inputs,
            requires_docker=False,
        )

    def _setup(self):
        # Check if x86_64 (aka AMD64) CPU is available locally
        machine = platform.uname().machine
        if machine != "x86_64" and machine != "AMD64":
            msg = "Only x86_64 and AMD64 CPUs are supported, got {machine}"
            raise exp.ModelRuntimeError(msg)

        self._transfer_files([self.conda_script])

    def _execute(
        self,
        output_dir: str,
        onnx_file: str,
        outputs_file: str,
    ):
        conda_env_name = "turnkey-onnxruntime-mlas-ep"

        try:
            # Create and setup the conda env
            create_conda_env(conda_env_name)
        except PermissionError as pe:
            os_type = platform.system()
            if os_type == "Windows":
                raise plugin_helpers.CondaError(
                    f"Conda environment setup encountered a permission issue: {pe}. "
                    "Ensure you have write permissions for the Conda installation directory. "
                    "If Conda is installed for 'All Users' on a Windows machine, it defaults "
                    "to 'C:\\ProgramData', where you may not have the necessary permissions."
                    "To resolve this, consider reinstalling Conda for 'Just Me' instead."
                )
            else:
                raise plugin_helpers.CondaError(
                    f"Conda env setup failed due to permission error: {pe}"
                )
        except Exception as e:
            raise plugin_helpers.CondaError(
                f"Conda env setup failed with exception: {e}"
            )

        # Execute the benchmark script in the conda environment
        execute_benchmark(
            onnx_file=onnx_file,
            outputs_file=outputs_file,
            output_dir=output_dir,
            conda_env_name=conda_env_name,
            iterations=self.iterations,
        )

    @property
    def mean_latency(self):
        return float(self._get_stat("Mean Latency(ms)"))

    @property
    def throughput(self):
        return float(self._get_stat("Throughput"))

    @property
    def device_name(self):
        return self._get_stat("CPU Name")



from typing import Optional, List, Tuple, Union, Dict, Any, Type, Callable
import sys
import os
import copy
import torch
import onnx
import turnkeyml.common.build as build
import turnkeyml.common.filesystem as filesystem
import turnkeyml.common.exceptions as exp
import turnkeyml.common.printing as printing
import turnkeyml.common.tf_helpers as tf_helpers
import turnkeyml.build.onnx_helpers as onnx_helpers
import turnkeyml.build.tensor_helpers as tensor_helpers
import turnkeyml.build.export as export
import turnkeyml.build.stage as stage
import turnkeyml.build.hummingbird as hummingbird
import turnkeyml.build.sequences as sequences
from turnkeyml.version import __version__ as turnkey_version


def lock_config(
    model: build.UnionValidModelInstanceTypes,
    build_name: Optional[str] = None,
    sequence: stage.Sequence = None,
    onnx_opset: Optional[int] = None,
    device: Optional[str] = None,
) -> build.Config:
    """
    Process the user's configuration arguments to build_model():
    1. Raise exceptions for illegal arguments
    2. Replace unset arguments with default values
    3. Lock the configuration into an immutable object
    """

    # The default model name is the name of the python file that calls build_model()
    auto_name = False
    if build_name is None:
        build_name = os.path.basename(sys.argv[0])
        auto_name = True

    if sequence is None:
        # The value ["default"] indicates that build_model() will be assigning some
        # default sequence later in the program
        stage_names = ["default"]
    else:
        stage_names = sequence.get_names()

    # Detect and validate ONNX opset
    if isinstance(model, str) and model.endswith(".onnx"):
        onnx_file_opset = onnx_helpers.get_opset(onnx.load(model))

        if onnx_opset is not None and onnx_opset != onnx_file_opset:
            raise ValueError(
                "When using a '.onnx' file as input, the onnx_opset argument must "
                "be None or exactly match the ONNX opset of the '.onnx' file. However, the "
                f"'.onnx' file has opset {onnx_file_opset}, while onnx_opset was set "
                f"to {onnx_opset}"
            )

        opset_to_use = onnx_file_opset
    else:
        if onnx_opset is None:
            opset_to_use = build.DEFAULT_ONNX_OPSET
        else:
            opset_to_use = onnx_opset

    if device is None:
        device_to_use = build.DEFAULT_DEVICE
    else:
        device_to_use = device

    # Store the args that should be immutable
    config = build.Config(
        build_name=build_name,
        auto_name=auto_name,
        sequence=stage_names,
        onnx_opset=opset_to_use,
        device=device_to_use,
    )

    return config


def decode_version_number(version: str) -> Dict[str, int]:
    numbers = [int(x) for x in version.split(".")]
    return {"major": numbers[0], "minor": numbers[1], "patch": numbers[0]}


def validate_cached_model(
    config: build.Config,
    model_type: build.ModelType,
    state: build.State,
    model: build.UnionValidModelInstanceTypes = None,
    inputs: Optional[Dict[str, Any]] = None,
) -> List[str]:
    """
    Verify whether anything in the call to build_model() changed
    We require the user to resolve the discrepancy when such a
    change occurs, so the purpose of this function is simply to
    detect these conditions and raise an appropriate error.
    If this function returns without raising an exception then
    the cached model is valid to use in the build.
    """

    result = []

    current_version_decoded = decode_version_number(turnkey_version)
    state_version_decoded = decode_version_number(state.turnkey_version)

    out_of_date: Union[str, bool] = False
    if current_version_decoded["major"] > state_version_decoded["major"]:
        out_of_date = "major"
    elif current_version_decoded["minor"] > state_version_decoded["minor"]:
        out_of_date = "minor"

    if out_of_date:
        msg = (
            f"Your build {state.config.build_name} was previously built against "
            f"turnkey version {state.turnkey_version}, "
            f"however you are now using onxxflow version {turnkey_version}. The previous build is "
            f"incompatible with this version of turnkey, as indicated by the {out_of_date} "
            "version number changing. See **docs/versioning.md** for details."
        )
        result.append(msg)

    if model is not None:
        model_changed = state.model_hash != build.hash_model(model, model_type)
    else:
        model_changed = False

    if inputs is not None:
        (
            input_shapes_changed,
            input_dtypes_changed,
        ) = tensor_helpers.check_shapes_and_dtypes(
            inputs,
            state.expected_input_shapes,
            state.expected_input_dtypes,
            expect_downcast=state.downcast_applied,
            raise_error=False,
        )
    else:
        input_shapes_changed = False
        input_dtypes_changed = False

    changed_args = []
    for key in vars(state.config):
        if vars(config)[key] != vars(state.config)[key]:
            changed_args.append((key, vars(config)[key], vars(state.config)[key]))

    # Show an error if the model changed
    build_conditions_changed = (
        model_changed
        or input_shapes_changed
        or input_dtypes_changed
        or len(changed_args) > 0
    )
    if build_conditions_changed:
        # Show an error if build_name is not specified for different models on the same script
        if state.uid == build.unique_id():
            msg = (
                "You are building multiple different models in the same script "
                "without specifying a unique build_model(..., build_name=) for each build."
            )
            result.append(msg)

        if model_changed:
            msg = (
                f'Model "{config.build_name}" changed since the last time it was built.'
            )
            result.append(msg)

        if input_shapes_changed:
            input_shapes, _ = build.get_shapes_and_dtypes(inputs)
            msg = (
                f'Input shape of model "{config.build_name}" changed from '
                f"{state.expected_input_shapes} to {input_shapes} "
                f"since the last time it was built."
            )
            result.append(msg)

        if input_dtypes_changed:
            _, input_dtypes = build.get_shapes_and_dtypes(inputs)
            msg = (
                f'Input data type of model "{config.build_name}" changed from '
                f"{state.expected_input_dtypes} to {input_dtypes} "
                f"since the last time it was built."
            )
            result.append(msg)

        if len(changed_args) > 0:
            for key_name, current_arg, previous_arg in changed_args:
                msg = (
                    f'build_model() argument "{key_name}" for build '
                    f"{config.build_name} changed from "
                    f"{previous_arg} to {current_arg} since the last build."
                )
                result.append(msg)
    else:
        if (
            state.build_status == build.FunctionStatus.ERROR
            or state.build_status == build.FunctionStatus.INCOMPLETE
            or state.build_status == build.FunctionStatus.KILLED
        ) and turnkey_version == state.turnkey_version:
            msg = (
                "build_model() has detected that you already attempted building "
                "this model with the exact same model, inputs, options, and version of "
                "turnkey, and that build failed."
            )
            result.append(msg)

    return result


def _begin_fresh_build(
    state_args: Dict,
    state_type: Type = build.State,
) -> build.State:
    # Wipe everything in this model's build directory, except for the stats file,
    # start with a fresh State.
    stats = filesystem.Stats(state_args["cache_dir"], state_args["config"].build_name)

    build_dir = build.output_dir(
        state_args["cache_dir"], state_args["config"].build_name
    )

    filesystem.rmdir(
        build_dir,
        excludes=[
            stats.file,
            os.path.join(build_dir, filesystem.BUILD_MARKER),
        ],
    )
    state = state_type(**state_args)
    state.save()

    return state


def _rebuild_if_needed(
    problem_report: str, state_args: Dict, state_type: Type = build.State
):
    build_name = state_args["config"].build_name
    msg = (
        f"build_model() discovered a cached build of {build_name}, but decided to "
        "rebuild for the following reasons: \n\n"
        f"{problem_report} \n\n"
        "build_model() will now rebuild your model to ensure correctness. You can change this "
        "policy by setting the build_model(rebuild=...) argument."
    )
    printing.log_warning(msg)

    return _begin_fresh_build(state_args, state_type=state_type)


def load_or_make_state(
    config: build.Config,
    evaluation_id: str,
    cache_dir: str,
    rebuild: str,
    model_type: build.ModelType,
    monitor: bool,
    model: build.UnionValidModelInstanceTypes = None,
    inputs: Optional[Dict[str, Any]] = None,
    state_type: Type = build.State,
    cache_validation_func: Callable = validate_cached_model,
    extra_state_args: Optional[Dict] = None,
) -> build.State:
    """
    Decide whether we can load the model from the model cache
    (return a valid State instance) or whether we need to rebuild it (return
    a new State instance).
    """

    # Put all the args for making a new State instance into a dict
    # to help the following code be cleaner
    state_args = {
        "model": model,
        "inputs": inputs,
        "monitor": monitor,
        "rebuild": rebuild,
        "evaluation_id": evaluation_id,
        "cache_dir": cache_dir,
        "config": config,
        "model_type": model_type,
    }

    # Ensure that `rebuild` has a valid value
    if rebuild not in build.REBUILD_OPTIONS:
        raise ValueError(
            f"Received `rebuild` argument with value {rebuild}, "
            f"however the only allowed values of `rebuild` are {build.REBUILD_OPTIONS}"
        )

    # Allow customizations of turnkey to supply additional args
    if extra_state_args is not None:
        state_args.update(extra_state_args)

    if rebuild == "always":
        return _begin_fresh_build(state_args, state_type)
    else:
        # Try to load state and check if model successfully built before
        if os.path.isfile(build.state_file(cache_dir, config.build_name)):
            try:
                state = build.load_state(
                    cache_dir,
                    config.build_name,
                    state_type=state_type,
                )

            except exp.StateError as e:
                problem = (
                    "- build_model() failed to load "
                    f"{build.state_file(cache_dir, config.build_name)}"
                )

                if rebuild == "if_needed":
                    return _rebuild_if_needed(problem, state_args, state_type)
                else:
                    # Give the rebuild="never" users a chance to address the problem
                    raise exp.CacheError(e)

            if (
                model_type == build.ModelType.UNKNOWN
                and state.build_status == build.FunctionStatus.SUCCESSFUL
            ):
                msg = (
                    "Model caching is disabled for successful builds against custom Sequences. "
                    "Your model will rebuild whenever you call build_model() on it."
                )
                printing.log_warning(msg)

                return _begin_fresh_build(state_args, state_type)
            elif (
                model_type == build.ModelType.UNKNOWN
                and state.build_status == build.FunctionStatus.INCOMPLETE
            ):
                msg = (
                    f"Model {config.build_name} was partially built in a previous call to "
                    "build_model(). This call to build_model() found that partial build and "
                    "is loading it from the build cache."
                )

                printing.log_info(msg)
            else:
                cache_problems = cache_validation_func(
                    config=config,
                    model_type=model_type,
                    state=state,
                    model=model,
                    inputs=inputs,
                )

                if len(cache_problems) > 0:
                    cache_problems = [f"- {msg}" for msg in cache_problems]
                    problem_report = "\n".join(cache_problems)

                    if rebuild == "if_needed":
                        return _rebuild_if_needed(
                            problem_report, state_args, state_type
                        )
                    if rebuild == "never":
                        msg = (
                            "build_model() discovered a cached build of "
                            f"{config.build_name}, and found that it "
                            "is likely invalid for the following reasons: \n\n"
                            f"{problem_report} \n\n"
                            "build_model() will raise a SkipBuild exception because you have "
                            "set rebuild=never. "
                        )
                        printing.log_warning(msg)

                        raise exp.SkipBuild(
                            "Skipping this build, by raising an exception, because it previously "
                            "failed and the `rebuild` argument is set to `never`."
                        )

            # Ensure the model and inputs are part of the state
            # This is useful  when loading models that still need to be built
            state.save_when_setting_attribute = False
            if state.model is None:
                state.model = model
            if state.inputs is None:
                state.inputs = inputs
            state.save_when_setting_attribute = True

            return state

        else:
            # No state file found, so we have to build
            return _begin_fresh_build(state_args, state_type)


export_map = {
    build.ModelType.PYTORCH: export.ExportPytorchModel(),
    build.ModelType.KERAS: export.ExportKerasModel(),
    build.ModelType.ONNX_FILE: export.ReceiveOnnxModel(),
    build.ModelType.HUMMINGBIRD: hummingbird.ConvertHummingbirdModel(),
}


def validate_inputs(inputs: Dict):
    """
    Check the model's inputs and make sure they are legal. Raise an exception
    if they are not legal.
    TODO: it may be wise to validate the inputs against the model, or at least
    the type of model, as well.
    """

    if inputs is None:
        msg = """
        build_model() requires model inputs. Check your call to build_model() to make sure
        you are passing the inputs argument.
        """
        raise exp.IntakeError(msg)

    if not isinstance(inputs, dict):
        msg = f"""
        The "inputs" argument to build_model() is required to be a dictionary, where the
        keys map to the named arguments in the model's forward function. The inputs
        received by build_model() were of type {type(inputs)}, not dict.
        """
        raise exp.IntakeError(msg)


def identify_model_type(model) -> build.ModelType:
    # Validate that the model's type is supported by build_model()
    # and assign a ModelType tag
    if isinstance(model, (torch.nn.Module, torch.jit.ScriptModule)):
        model_type = build.ModelType.PYTORCH
    elif isinstance(model, str):
        if model.endswith(".onnx"):
            model_type = build.ModelType.ONNX_FILE
    elif tf_helpers.is_keras_model(model):
        model_type = build.ModelType.KERAS
        if not tf_helpers.is_executing_eagerly():
            raise exp.IntakeError(
                "`build_model()` requires Keras models to be run in eager execution mode. "
                "Enable eager execution to continue."
            )
        if not model.built:
            raise exp.IntakeError(
                "Keras model has not been built. Please call "
                "model.build(input_shape) before running build_model()"
            )
    elif hummingbird.is_supported_model(model):
        model_type = build.ModelType.HUMMINGBIRD
    else:
        raise exp.IntakeError(
            "Argument 'model' passed to build_model() is "
            f"of unsupported type {type(model)}"
        )

    return model_type


def model_intake(
    user_model,
    user_inputs,
    user_sequence: Optional[stage.Sequence],
) -> Tuple[Any, Any, stage.Sequence, build.ModelType, str]:
    # Model intake structure options:
    # user_model
    #    |
    #    |------- path to onnx model file
    #    |
    #    |------- pytorch model object
    #    |
    #    |------- keras model object
    #    |
    #    |------- Hummingbird-supported model object

    if user_sequence is None or user_sequence.enable_model_validation:
        if user_model is None and user_inputs is None:
            msg = """
            You are running build_model() without any model, inputs, or custom Sequence. The purpose
            of non-customized build_model() is to build a model against some inputs, so you need to
            provide both.
            """
            raise exp.IntakeError(msg)

        # Make sure that if the model is a file path, it is valid
        if isinstance(user_model, str):
            if not os.path.isfile(user_model):
                msg = f"""
                build_model() model argument was passed a string (path to a model file),
                however no file was found at {user_model}.
                """
                raise exp.IntakeError(msg)

            if not user_model.endswith(".onnx"):
                msg = f"""
                build_model() received a model argument that was a string. However, model string
                arguments are required to be a path to a .onnx file, but the argument was: {user_model}
                """
                raise exp.IntakeError(msg)

            # Create dummy inputs based on the ONNX spec, if none were provided by the user
            if user_inputs is None:
                inputs = onnx_helpers.dummy_inputs(user_model)
            else:
                inputs = user_inputs
        else:
            inputs = user_inputs

        model_type = identify_model_type(user_model)

        sequence = copy.deepcopy(user_sequence)
        if sequence is None:
            sequence = stage.Sequence(
                "top_level_sequence",
                "Top Level Sequence",
                [sequences.onnx_fp32],
            )

        # If there is an ExportPlaceholder Stage in the sequence, replace it with
        # a framework-specific export Stage.
        # First, make a deepcopy of any sequence we bring in here. We do not want to modify
        # the original.
        sequence = copy.deepcopy(sequence)
        for index, stage_instance in enumerate(sequence.stages):
            if isinstance(stage_instance, export.ExportPlaceholder):
                sequence.stages[index] = export_map[model_type]

        validate_inputs(inputs)

    else:
        # We turn off a significant amount of automation and validation
        # to provide custom stages and sequences with maximum flexibility
        inputs = user_inputs
        sequence = user_sequence
        model_type = build.ModelType.UNKNOWN

    return (user_model, inputs, sequence, model_type)



# labels: name::bartpho_word author::transformers task::Generative_AI license::mit
from turnkeyml.parser import parse
from transformers import AutoModel, AutoTokenizer
import torch

torch.manual_seed(0)

# Parsing command-line arguments
pretrained, batch_size, max_seq_length = parse(
    ["pretrained", "batch_size", "max_seq_length"]
)

# Model and input configurations
if pretrained:
    tokenizer = AutoTokenizer.from_pretrained("vinai/bartpho-word")
    model = AutoModel.from_pretrained("vinai/bartpho-word")
else:
    raise ValueError(
        "This model is only supported with pretrained weights, try again with --pretrained"
    )

# Make sure the user's sequence length fits within the model's maximum
assert max_seq_length <= model.config.max_position_embeddings


# Subtract 2 for the start and end token
text_sentence = "word " * (max_seq_length - 2)
inputs = {
    "input_ids": torch.tensor(
        tokenizer(text_sentence, return_tensors="pt")["input_ids"] * batch_size
    )
}


# Call model
model(**inputs)



import time
import os
import copy
import glob
import pathlib
from typing import Tuple, List, Dict, Optional, Union
import turnkeyml.common.printing as printing
import turnkeyml.common.exceptions as exceptions
import turnkeyml.build.stage as stage
import turnkeyml.cli.spawn as spawn
import turnkeyml.common.filesystem as filesystem
import turnkeyml.common.labels as labels_library
import turnkeyml.run.devices as devices
from turnkeyml.common.performance import Device
from turnkeyml.run.devices import SUPPORTED_RUNTIMES
from turnkeyml.analyze.script import (
    evaluate_script,
    TracerArgs,
    Action,
    explore_invocation,
    get_model_hash,
)
from turnkeyml.analyze.util import ModelInfo, UniqueInvocationInfo
import turnkeyml.common.build as build
import turnkeyml.build.onnx_helpers as onnx_helpers


def decode_input_arg(input: str) -> Tuple[str, List[str], str]:
    # Parse the targets out of the file name
    # Targets use the format:
    #   file_path.ext::target0,target1,...,targetN
    decoded_input = input.split("::")
    file_path = os.path.abspath(decoded_input[0])

    if len(decoded_input) == 2:
        targets = decoded_input[1].split(",")
        encoded_input = file_path + "::" + decoded_input[1]
    elif len(decoded_input) == 1:
        targets = []
        encoded_input = file_path
    else:
        raise ValueError(
            "Each file input to turnkey should have either 0 or 1 '::' in it."
            f"However, {file_path} was received."
        )

    return file_path, targets, encoded_input


def check_sequence_type(
    sequence: Union[str, stage.Sequence],
    use_slurm: bool,
    process_isolation: bool,
):
    """
    Check to make sure the user's sequence argument is valid.
    use_slurm or process_isolation: only work with names of installed sequences
    otherwise: sequence instances and sequence names are allowed
    """

    if sequence is not None:
        if use_slurm or process_isolation:
            # The spawned process will need to load a sequence file
            if not isinstance(sequence, str):
                raise ValueError(
                    "The 'sequence' arg must be a str (name of an installed sequence) "
                    "when use_slurm=True or process_isolation=True."
                )


def unpack_txt_inputs(input_files: List[str]) -> List[str]:
    """
    Replace txt inputs with models listed inside those files
    Note: This implementation allows for nested .txt files
    """
    txt_files_expanded = sum(
        [glob.glob(f) for f in input_files if f.endswith(".txt")], []
    )
    processed_files = []
    for input_string in txt_files_expanded:
        if not os.path.exists(input_string):
            raise exceptions.ArgError(
                f"{input_string} does not exist. Please verify the file."
            )

        with open(input_string, "r", encoding="utf-8") as file:
            lines = [line.strip() for line in file if line.strip() != ""]
            processed_files.extend(unpack_txt_inputs(lines))

    return processed_files + [f for f in input_files if not f.endswith(".txt")]


# pylint: disable=unused-argument
def benchmark_files(
    input_files: List[str],
    use_slurm: bool = False,
    process_isolation: bool = False,
    lean_cache: bool = False,
    cache_dir: str = filesystem.DEFAULT_CACHE_DIR,
    labels: List[str] = None,
    rebuild: Optional[str] = None,
    device: str = "x86",
    runtime: str = None,
    iterations: int = 100,
    analyze_only: bool = False,
    build_only: bool = False,
    script_args: Optional[str] = None,
    max_depth: int = 0,
    onnx_opset: Optional[int] = None,
    timeout: Optional[int] = None,
    sequence: Union[str, stage.Sequence] = None,
    rt_args: Optional[Dict] = None,
):
    # Capture the function arguments so that we can forward them
    # to downstream APIs
    benchmarking_args = copy.deepcopy(locals())
    regular_files = []

    # Replace .txt files with the models listed inside them
    input_files = unpack_txt_inputs(input_files)

    # Iterate through each string in the input_files list
    for input_string in input_files:
        if not any(char in input_string for char in "*?[]"):
            regular_files.append(input_string)

    # Create a list of files that don't exist on the filesystem
    # Skip the files with "::" as hashes will be decoded later
    non_existent_files = [
        file for file in regular_files if not os.path.exists(file) and "::" not in file
    ]
    if non_existent_files:
        raise exceptions.ArgError(
            f"{non_existent_files} do not exist, please verify if the file(s) exists."
        )

    # Make sure that `timeout` is only being used with `process_isolation` or `use_slurm`
    # And then set a default for timeout if the user didn't set a value
    if timeout is not None:
        if not use_slurm and not process_isolation:
            raise exceptions.ArgError(
                "The `timeout` argument is only allowed when slurm "
                "or process isolation mode is activated."
            )

        timeout_to_use = timeout
    else:
        timeout_to_use = spawn.DEFAULT_TIMEOUT_SECONDS

    benchmarking_args["timeout"] = timeout_to_use

    # Convert regular expressions in input files argument
    # into full file paths (e.g., [*.py] -> [a.py, b.py] )
    input_files_expanded = filesystem.expand_inputs(input_files)

    # Do not forward arguments to downstream APIs
    # that will be decoded in this function body
    benchmarking_args.pop("input_files")
    benchmarking_args.pop("labels")
    benchmarking_args.pop("use_slurm")
    benchmarking_args.pop("process_isolation")

    # Make sure the cache directory exists
    filesystem.make_cache_dir(cache_dir)

    check_sequence_type(sequence, use_slurm, process_isolation)

    if device is None:
        device = "x86"

    # Replace the runtime with a default value, if needed
    selected_runtime = devices.apply_default_runtime(device, runtime)
    benchmarking_args["runtime"] = selected_runtime

    # Get the default part and config by providing the Device class with
    # the supported devices by the runtime
    runtime_supported_devices = SUPPORTED_RUNTIMES[selected_runtime][
        "supported_devices"
    ]
    benchmarking_args["device"] = str(Device(device, runtime_supported_devices))

    # Force the user to specify a legal cache dir in NFS if they are using slurm
    if cache_dir == filesystem.DEFAULT_CACHE_DIR and use_slurm:
        printing.log_warning(
            "Using the default cache directory when using Slurm will cause your cached "
            "files to only be available at the Slurm node. If this is not the behavior "
            "you desired, please se a --cache-dir that is accessible by both the slurm "
            "node and your local machine."
        )

    # Get list containing only file names
    clean_file_names = [
        decode_input_arg(file_name)[0] for file_name in input_files_expanded
    ]

    # Validate that the files have supported file extensions
    # Note: We are not checking for .txt files here as those were previously handled
    for file_name in clean_file_names:
        if not file_name.endswith(".py") and not file_name.endswith(".onnx"):
            raise exceptions.ArgError(
                f"File extension must be .py, .onnx, or .txt (got {file_name})"
            )

    # Decode turnkey args into TracerArgs flags
    if analyze_only:
        actions = [
            Action.ANALYZE,
        ]
    elif build_only:
        actions = [
            Action.ANALYZE,
            Action.BUILD,
        ]
    else:
        actions = [
            Action.ANALYZE,
            Action.BUILD,
            Action.BENCHMARK,
        ]

    if use_slurm:
        jobs = spawn.slurm_jobs_in_queue()
        if len(jobs) > 0:
            printing.log_warning(f"There are already slurm jobs in your queue: {jobs}")
            printing.log_info(
                "Suggest quitting turnkey, running 'scancel -u $USER' and trying again."
            )

    # Use this data structure to keep a running index of all models
    models_found: Dict[str, ModelInfo] = {}

    # Fork the args for analysis since they have differences from the spawn args:
    # build_only and analyze_only are encoded into actions
    analysis_args = copy.deepcopy(benchmarking_args)
    analysis_args.pop("build_only")
    analysis_args.pop("analyze_only")
    analysis_args["actions"] = actions
    analysis_args.pop("timeout")

    for file_path_encoded in input_files_expanded:
        # Check runtime requirements if needed. All benchmarking will be halted
        # if requirements are not met. This happens regardless of whether
        # process-isolation is used or not.
        runtime_info = SUPPORTED_RUNTIMES[selected_runtime]
        if "requirement_check" in runtime_info and Action.BENCHMARK in actions:
            runtime_info["requirement_check"]()

        printing.log_info(f"Running turnkey on {file_path_encoded}")

        file_path_absolute, targets, encoded_input = decode_input_arg(file_path_encoded)

        # Skip a file if the required_labels are not a subset of the script_labels.
        if labels:
            # Labels argument is not supported for ONNX files
            if file_path_absolute.endswith(".onnx"):
                raise ValueError(
                    "The labels argument is not supported for .onnx files, got",
                    file_path_absolute,
                )
            required_labels = labels_library.to_dict(labels)
            script_labels = labels_library.load_from_file(encoded_input)
            if not labels_library.is_subset(required_labels, script_labels):
                continue

        if use_slurm or process_isolation:
            # Decode args into spawn.Target
            if use_slurm and process_isolation:
                raise ValueError(
                    "use_slurm and process_isolation are mutually exclusive, but both are True"
                )
            elif use_slurm:
                process_type = spawn.Target.SLURM
            elif process_isolation:
                process_type = spawn.Target.LOCAL_PROCESS
            else:
                raise ValueError(
                    "This code path requires use_slurm or use_process to be True, "
                    "but both are False"
                )

            spawn.run_turnkey(
                op="benchmark",
                target=process_type,
                file_name=encoded_input,
                **benchmarking_args,
            )

        else:
            # Instantiate an object that holds all of the arguments
            # for analysis, build, and benchmarking
            tracer_args = TracerArgs(
                models_found=models_found,
                targets=targets,
                input=file_path_absolute,
                **analysis_args,
            )

            if file_path_absolute.endswith(".py"):
                # Run analysis, build, and benchmarking on every model
                # in the python script
                models_found = evaluate_script(tracer_args)
            elif file_path_absolute.endswith(".onnx"):
                # Skip analysis and go straight to dealing with the model
                # We need to manufacture ModelInfo and UniqueInvocatioInfo instances to do this,
                # since we didn't get them from analysis.

                # Gather information about the ONNX model
                onnx_name = pathlib.Path(file_path_absolute).stem
                onnx_hash = get_model_hash(
                    file_path_absolute, build.ModelType.ONNX_FILE
                )
                onnx_inputs = onnx_helpers.dummy_inputs(file_path_absolute)
                input_shapes = {key: value.shape for key, value in onnx_inputs.items()}

                # Create the UniqueInvocationInfo
                #  - execute=1 is required or else the ONNX model will be
                #       skipped in later stages of evaluation
                #  - is_target=True is required or else traceback wont be printed for
                #       in the event of any errors
                #  - Most other values can be left as default
                invocation_info = UniqueInvocationInfo(
                    executed=1,
                    input_shapes=input_shapes,
                    hash=onnx_hash,
                    is_target=True,
                )

                # Create the ModelInfo
                model_info = ModelInfo(
                    model=file_path_absolute,
                    name=onnx_name,
                    script_name=onnx_name,
                    file=file_path_absolute,
                    build_model=not build_only,
                    model_type=build.ModelType.ONNX_FILE,
                    unique_invocations={onnx_hash: invocation_info},
                    hash=onnx_hash,
                )

                # Begin evaluating the ONNX model
                tracer_args.script_name = onnx_name
                tracer_args.models_found[tracer_args.script_name] = model_info
                explore_invocation(
                    model_inputs=onnx_inputs,
                    model_info=model_info,
                    invocation_info=invocation_info,
                    tracer_args=tracer_args,
                )
                models_found = tracer_args.models_found

    # Wait until all the Slurm jobs are done
    if use_slurm:
        while len(spawn.slurm_jobs_in_queue()) != 0:
            print(
                f"Waiting: {len(spawn.slurm_jobs_in_queue())} "
                f"jobs left in queue: {spawn.slurm_jobs_in_queue()}"
            )
            time.sleep(5)

    printing.log_success("The 'benchmark' command is complete.")



import turnkeyml.build.sequences as sequences
from .runtime import TensorRT


implements = {
    "runtimes": {
        "trt": {
            "build_required": True,
            "RuntimeClass": TensorRT,
            "supported_devices": {"nvidia"},
            "default_sequence": sequences.optimize_fp16,
        }
    }
}



# labels: test_group::turnkey name::pnaconv author::graph_convolutions task::Graph_Machine_Learning license::mit
"""
The Principal Neighbourhood Aggregation graph convolution operator from the
`"Principal Neighbourhood Aggregation for Graph Nets"
<https://arxiv.org/abs/2004.05718>`_ paper
"""
import torch

from torch_geometric.datasets import Planetoid
from torch_geometric.nn import PNAConv

dataset = Planetoid(root=".", name="Cora")
data = dataset[0]
edge_index_rows = 2
deg = torch.ones(1, data.num_nodes, dtype=torch.int)


# PNAConv(in_channels, out_channels, aggregators, scalers, deg)
model = PNAConv(dataset.num_features, dataset.num_classes, ["sum"], ["identity"], deg)
x = torch.ones(data.num_nodes, data.num_features, dtype=torch.float)
edge_index = torch.ones(edge_index_rows, data.num_nodes, dtype=torch.long)
inputs = {
    "x": x,
    "edge_index": edge_index,
}

# Call model
model(**inputs)



import os
import argparse


def login(key: str):

    toke_path = os.path.expanduser("~/.huggingface/token")
    hf_path = os.path.expanduser("~/.huggingface")
    if not os.path.exists(hf_path):
        os.makedirs(hf_path)

    if not os.path.exists(toke_path):
        with open(toke_path, "w", encoding="utf-8") as f:
            f.write(key)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Update huggingface credentials")
    parser.add_argument("--key", help="API key", required=True)
    args = vars(parser.parse_args())
    login(args["key"])



# labels: name::twolayer author::selftest
import torch

torch.manual_seed(0)


class TwoLayerTestModel(torch.nn.Module):
    def __init__(self, input_size, output_size):
        super(TwoLayerTestModel, self).__init__()
        self.fc = torch.nn.Linear(input_size, output_size)
        self.fc2 = torch.nn.Linear(output_size, output_size)

    def forward(self, x):
        output = self.fc(x)
        output = self.fc2(output)
        return output


# Instantiate model and generate inputs
input_features = 10
output_features = 5

# Model and input configurations
model = TwoLayerTestModel(input_features, output_features)
inputs = {"x": torch.rand(input_features)}

# Call model
model(**inputs)



# labels: test_group::turnkey name::fcos_resnet50_fpn author::torchvision task::Computer_Vision license::bsd-3-clause
"""
https://pytorch.org/vision/stable/models/fcos.html
"""

from turnkeyml.parser import parse
import torch
from torchvision.models.detection import fcos_resnet50_fpn, FCOS_ResNet50_FPN_Weights


torch.manual_seed(0)

# Parsing command-line arguments
pretrained, batch_size, num_channels, width, height = parse(
    ["pretrained", "batch_size", "num_channels", "width", "height"]
)


# Model and input configurations
model = fcos_resnet50_fpn(
    weights=FCOS_ResNet50_FPN_Weights.DEFAULT if pretrained else None
)
model.eval()
inputs = {"images": torch.ones([batch_size, num_channels, width, height])}


# Call model
model(**inputs)



import os
import inspect
import shutil
import warnings
import sys
import copy
from typing import Union
import torch
import torch.onnx.verification
import numpy as np
import onnxruntime
import onnxmltools
import onnx
import turnkeyml.build.stage as stage
import turnkeyml.common.exceptions as exp
import turnkeyml.common.build as build
import turnkeyml.build.tensor_helpers as tensor_helpers
import turnkeyml.build.onnx_helpers as onnx_helpers
import turnkeyml.common.filesystem as fs


def check_model(onnx_file, success_message, fail_message) -> bool:
    if os.path.isfile(onnx_file):
        print(success_message)
    else:
        print(fail_message)
        return False
    try:
        onnx.checker.check_model(onnx_file)
        print("\tSuccessfully checked onnx file")
        return True
    except onnx.checker.ValidationError as e:
        print("\tError while checking generated ONNX file")
        print(e)
        return False


def _warn_to_stdout(message, category, filename, line_number, _, line):
    sys.stdout.write(
        warnings.formatwarning(message, category, filename, line_number, line)
    )


def get_output_names(
    onnx_model: Union[str, onnx.ModelProto]
):  # pylint: disable=no-member
    # Get output names of ONNX file/model
    if not isinstance(onnx_model, onnx.ModelProto):  # pylint: disable=no-member
        onnx_model = onnx.load(onnx_model)
    return [node.name for node in onnx_model.graph.output]  # pylint: disable=no-member


def onnx_dir(state: build.State):
    return os.path.join(
        build.output_dir(state.cache_dir, state.config.build_name), "onnx"
    )


def base_onnx_file(state: build.State):
    return os.path.join(
        onnx_dir(state),
        f"{state.config.build_name}-op{state.config.onnx_opset}-base.onnx",
    )


def opt_onnx_file(state: build.State):
    return os.path.join(
        onnx_dir(state),
        f"{state.config.build_name}-op{state.config.onnx_opset}-opt.onnx",
    )


def converted_onnx_file(state: build.State):
    return os.path.join(
        onnx_dir(state),
        f"{state.config.build_name}-op{state.config.onnx_opset}-opt-f16.onnx",
    )


class ExportPlaceholder(stage.Stage):
    """
    Placeholder Stage that should be replaced by a framework-specific export stage,
    typically during ignition.model_intake()
    """

    def __init__(self):
        super().__init__(
            unique_name="export_placeholder",
            monitor_message="Placeholder for an Export Stage",
        )

    def fire(self, _: build.State):
        raise exp.StageError(
            "This Sequence includes an ExportPlaceholder Stage that should have "
            "been replaced with an export Stage."
        )


class ReceiveOnnxModel(stage.Stage):
    """
    Stage that takes an ONNX model as input.

    Expected inputs:
     - state.model is a path to the ONNX model
     - state.inputs is a dict that represents valid inputs for the onnx model

    Outputs:
     - A *-base.onnx file that implements state.model given state.inputs.
    """

    def __init__(self):
        super().__init__(
            unique_name="receive_onnx",
            monitor_message="Receiving ONNX Model",
        )

    def fire(self, state: build.State):
        if not isinstance(state.model, str):
            msg = f"""
            The current stage (ReceiveOnnxModel) is only compatible with
            ONNX files, however the stage received a model of type
            {type(state.model)}.
            """
            raise exp.StageError(msg)
        if not state.model.endswith(".onnx"):
            msg = f"""
            The current stage (ReceiveOnnxModel) expects a path to ONNX
            model, however the stage received {state.model}.
            """
            raise exp.StageError(msg)

        dummy_inputs = tuple(state.inputs.values())
        dummy_input_names = tuple(state.inputs.keys())
        state.inputs = dict(zip(dummy_input_names, dummy_inputs))

        model = onnx.load(state.model)
        opset = onnx_helpers.get_opset(model)
        input_shapes = [
            [d.dim_value for d in _input.type.tensor_type.shape.dim]
            for _input in model.graph.input  # pylint: disable=no-member
        ]

        # Save output node names
        state.expected_output_names = get_output_names(model)

        # Check for Dynamic shapes in the model. They can be represented as 0, -1, "unk__".
        for input in input_shapes:
            for dimension in input:
                if dimension < 1 or not isinstance(dimension, int):
                    msg = f"""
                    The received model has dynamic input dimensions. Please freeze the model with static
                    input dimensions.
                    More information may be available in the log file at **{self.logfile_path}**
                    """
                    raise exp.StageError(msg)

        if opset < build.DEFAULT_ONNX_OPSET and opset >= build.MINIMUM_ONNX_OPSET:
            print(
                f" \n The received model has an opset {opset}. Though this opset is supported \
                we recommend upgrading the model to opset {build.MINIMUM_ONNX_OPSET}"
            )
        elif opset < build.MINIMUM_ONNX_OPSET:
            msg = f"""
            The received model has an opset {opset}. Opset < 11 is not supported. Please
            try upgrading the model to opset 13.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        output_path = base_onnx_file(state)
        os.makedirs(onnx_dir(state), exist_ok=True)
        shutil.copy(state.model, output_path)

        tensor_helpers.save_inputs(
            [state.inputs], state.original_inputs_file, downcast=False
        )

        # Check the if the base mode has been exported successfully
        success_msg = "\tSuccess receiving ONNX Model"
        fail_msg = "\tFailed receiving ONNX Model"

        if check_model(output_path, success_msg, fail_msg):
            state.intermediate_results = [output_path]

            stats = fs.Stats(
                state.cache_dir, state.config.build_name, state.evaluation_id
            )
            stats.save_model_eval_stat(
                fs.Keys.ONNX_FILE,
                output_path,
            )
        else:
            msg = f"""
            Unable to process ONNX Model. We recommend that you verify the source of the model.
            Any optimizations performed on the model could result in an error.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        return state


class ExportPytorchModel(stage.Stage):
    """
    Stage that takes a PyTorch model instance, in state.model, and
    exports it to an ONNX file.

    Expected inputs:
     - state.model is a torch.nn.Module or torch.jit.ScriptModule
     - state.inputs is a dict that represents valid kwargs to the forward
        function of state.model

    Outputs:
     - A *-base.onnx file that implements state.model given state.inputs
    """

    def __init__(self):
        super().__init__(
            unique_name="export_pytorch",
            monitor_message="Exporting PyTorch to ONNX",
        )

    def fire(self, state: build.State):
        if not isinstance(state.model, (torch.nn.Module, torch.jit.ScriptModule)):
            msg = f"""
            The current stage (ExportPytorchModel) is only compatible with
            models of type torch.nn.Module or torch.jit.ScriptModule, however
            the stage received a model of type {type(state.model)}.
            """
            raise exp.StageError(msg)

        # The `torch.onnx.export()` function accepts a tuple of positional inputs
        # followed by a dictionary with all keyword inputs.
        # The dictionary must be last item in tuple.
        user_provided_args = list(state.inputs.keys())

        if isinstance(state.model, torch.nn.Module):
            # Validate user provided args
            all_args = list(inspect.signature(state.model.forward).parameters.keys())

            for inp in user_provided_args:
                if inp not in all_args:
                    msg = f"""
                    Input name {inp} not found in the model's forward method. Available
                    input names are: {all_args}"
                    """
                    raise ValueError(msg)

            # Most pytorch models have args that are kind = positional_or_keyword.
            # The `torch.onnx.export()` function accepts model args as
            #     (all_positional_args_value,{keyword_arg:value}).
            # To map the input_args correctly and to build an accurate model
            # the order of the input_names must reflect the order of the model args.

            # Collect order of pytorch model args.
            all_args_order_mapping = {arg: idx for idx, arg in enumerate(all_args)}

            # Sort the user provided inputs with respect to model args and store as tuple.
            sorted_user_inputs = sorted(
                user_provided_args, key=lambda x: all_args_order_mapping[x]
            )
            dummy_input_names = tuple(sorted_user_inputs)

            # If a single input is provided torch.onnx.export will
            # not accept a dictionary, so pop the first arg
            user_args = copy.deepcopy(state.inputs)
            first_input = user_args.pop(dummy_input_names[0])

            # Create tuple: (first input, {rest of user_args dict as keyword args})
            dummy_inputs = (first_input, user_args)

        else:  # state.model is a torch.jit.ScriptModule
            dummy_inputs = tuple(state.inputs.values())

            # Collect input names
            dummy_input_names = tuple(state.inputs.keys())

        # Send torch export warnings to stdout (and therefore the log file)
        # so that they don't fill up the command line
        default_warnings = warnings.showwarning
        warnings.showwarning = _warn_to_stdout

        stats = fs.Stats(state.cache_dir, state.config.build_name, state.evaluation_id)

        # Verify if the exported model matches the input torch model
        try:
            # Tolerance levels for the torch export are recommended by Pytorch here:
            # https://pytorch.org/docs/stable/testing.html#module-torch.testing
            fp32_tolerance = torch.onnx.verification.VerificationOptions(
                rtol=1.3e-6, atol=1e-5
            )

            # The `torch.onnx.verification.find_mismatch()` takes input arguments to the
            # model as `input_args (Tuple[Any, ...])`
            export_verification = torch.onnx.verification.find_mismatch(
                state.model,
                tuple(state.inputs.values()),
                opset_version=state.config.onnx_opset,
                options=fp32_tolerance,
            )

            # `export_verification.has_mismatch()` returns True if a mismatch is found and
            # False otherwise. If no mismatch is found,# `is_export_valid` is set to "Valid",
            # indicating successful verification.
            # If a mismatch is found, `is_export_valid` is set to "Invalid", indicating
            # the verification failed.
            if not export_verification.has_mismatch():
                is_export_valid = "valid"
            else:
                is_export_valid = "invalid"

        # The except block catches any type of exception that might occur during the
        # verification process. If any exception occurs,`is_export_valid` is set to
        # "Unverified", indicating that the verification process could not be completed,
        # and therefore the model's export status is unverified.
        except Exception:  # pylint: disable=broad-except
            is_export_valid = "unverified"

        stats.save_model_eval_stat(
            fs.Keys.TORCH_ONNX_EXPORT_VALIDITY,
            is_export_valid,
        )

        # Export the model to ONNX
        output_path = base_onnx_file(state)
        os.makedirs(onnx_dir(state), exist_ok=True)

        torch.onnx.export(
            state.model,
            dummy_inputs,
            output_path,
            input_names=dummy_input_names,
            do_constant_folding=True,
            opset_version=state.config.onnx_opset,
            verbose=False,
        )

        # Save output names to ensure we are preserving the order of the outputs
        state.expected_output_names = get_output_names(output_path)

        # Restore default warnings behavior
        warnings.showwarning = default_warnings

        tensor_helpers.save_inputs(
            [state.inputs], state.original_inputs_file, downcast=False
        )

        # Check the if the base mode has been exported successfully
        success_msg = "\tSuccess exporting model to ONNX"
        fail_msg = "\tFailed exporting model to ONNX"

        if check_model(output_path, success_msg, fail_msg):
            state.intermediate_results = [output_path]

            stats.save_model_eval_stat(
                fs.Keys.ONNX_FILE,
                output_path,
            )
        else:
            msg = f"""
            Unable to export model to ONNX using Torch's ONNX exporter.
            We recommend that you modify your model until it is
            compatible with this third party software, then re-run.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        return state


class ExportKerasModel(stage.Stage):
    """
    Stage that takes a Keras model instance, in state.model, and
    exports it to an ONNX file.

    Expected inputs:
     - state.model is a tf.keras.Model
     - state.inputs is a dict that represents valid kwargs to the forward
        function of state.model

    Outputs:
     - A *-base.onnx file that implements state.model given state.inputs
    """

    def __init__(self):
        super().__init__(
            unique_name="export_keras",
            monitor_message="Exporting Keras to ONNX",
        )

    def fire(self, state: build.State):
        # pylint: disable=import-error
        import tensorflow as tf
        import tf2onnx

        if not isinstance(state.model, (tf.keras.Model)):
            msg = f"""
            The current stage (ExportKerasModel) is only compatible with
            models of type tf.keras.Model, however
            the stage received a model of type {type(state.model)}.
            """
            raise exp.StageError(msg)

        user_provided_args = state.inputs.keys()

        all_args = []

        # Check the model inputs member
        if state.model.inputs:
            all_args = [x.name for x in state.model.inputs]

        # If the input name(s) cannot be extracted from the inputs variable
        # than try to find them in the call() method
        if len(all_args) == 0:
            all_args = list(inspect.signature(state.model.call).parameters.keys())

        inputs = []
        input_names = []

        for inp in user_provided_args:
            if inp not in all_args:
                msg = f"""
                Input name {inp} not found in the model's forward method. Available
                input names are: {all_args}"
                """
                raise ValueError(msg)

        for _, arg in enumerate(all_args):
            if arg in user_provided_args:
                inputs.append(state.inputs[arg])
                input_names.append(arg)

        input_specs = []
        for inp, name in zip(inputs, input_names):
            dtype = inp.dtype
            shape = inp.shape
            if inp.dtype == tf.float64:
                print(f"Converting input {name} from float64 to float32")
                dtype = tf.float32
            if inp.dtype == tf.int64:
                print(f"Converting input {name} from int64 to int32")
                dtype = tf.int32
            if inp.shape[0] is None:
                print("Found batch size None and setting it to 1")
                shape = (1, shape[1:])

            input_specs.append(tf.TensorSpec(shape, dtype, name))

        # Export the model to ONNX
        output_path = base_onnx_file(state)
        os.makedirs(onnx_dir(state), exist_ok=True)
        tf2onnx.convert.from_keras(
            state.model,
            input_signature=input_specs,
            opset=state.config.onnx_opset,
            output_path=output_path,
        )

        # Save output names to ensure we are preserving the order of the outputs
        state.expected_output_names = get_output_names(output_path)

        state.inputs = dict(zip(tuple(input_names), tuple(inputs)))

        tensor_helpers.save_inputs(
            [state.inputs], state.original_inputs_file, downcast=False
        )

        # Check the if the base mode has been exported successfully
        success_msg = "\tSuccess exporting model to ONNX"
        fail_msg = "\tFailed exporting model to ONNX"

        if check_model(output_path, success_msg, fail_msg):
            state.intermediate_results = [output_path]

            stats = fs.Stats(
                state.cache_dir, state.config.build_name, state.evaluation_id
            )
            stats.save_model_eval_stat(
                fs.Keys.ONNX_FILE,
                output_path,
            )
        else:
            msg = f"""
            Unable to export model to ONNX using tf2onnx exporter.
            We recommend that you modify your model until it is
            compatible with this third party software, then re-run.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        return state


class OptimizeOnnxModel(stage.Stage):
    """
    Stage that takes an ONNX file and uses ONNX Runtime to optimize it.
    Important because this helps to perform constant folding, Redundant
    node eliminations, Semantics-preserving node fusions

    Expected inputs:
     - state.intermediate_results contains a single .onnx file

    Outputs:
     - A *-opt.onnx file
    """

    def __init__(self):
        super().__init__(
            unique_name="optimize_onnx",
            monitor_message="Optimizing ONNX file",
        )

    def fire(self, state: build.State):
        input_onnx = state.intermediate_results[0]
        output_path = opt_onnx_file(state)

        # Perform some basic optimizations on the model to remove shape related
        # information inserted for dynamic shape inference.
        # Given that we're compiling against a fixed sequence length the dynamic
        # shape information is not necessary
        session_options = onnxruntime.SessionOptions()

        # Set graph optimization level
        session_options.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_BASIC
        )

        # To enable model serialization after graph optimization set this
        session_options.optimized_model_filepath = output_path

        # Optimize graph
        onnxruntime.InferenceSession(input_onnx, session_options)

        # Check that the converted model is still valid
        success_msg = "\tSuccess optimizing ONNX model"
        fail_msg = "\tFailed optimizing ONNX model"

        if check_model(output_path, success_msg, fail_msg):
            state.intermediate_results = [output_path]

            stats = fs.Stats(
                state.cache_dir, state.config.build_name, state.evaluation_id
            )
            stats.save_model_eval_stat(
                fs.Keys.ONNX_FILE,
                output_path,
            )
        else:
            msg = f"""
            Unable to optimize ONNX file using ONNX runtime.
            We recommend that you modify your model until it is
            compatible with this third party software, then re-run.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        return state


class ConvertOnnxToFp16(stage.Stage):
    """
    Stage that takes an ONNX file and converts its trained parameters
    to fp16.

    Expected inputs:
     - state.intermediate_results contains a single .onnx file

    Outputs:
     - A *-f16.onnx file with FP16 trained parameters
    """

    def __init__(self):
        super().__init__(
            unique_name="fp16_conversion",
            monitor_message="Converting to FP16",
        )

    def fire(self, state: build.State):
        input_onnx = state.intermediate_results[0]

        # Convert the model to FP16
        # Some ops will not be converted to fp16 because they are in a block list
        # The latest list can be found here. It is not necessarily the list that
        # our version of onnxmltools sees
        # https://github.com/microsoft/onnxconverter-common/blob/master/onnxconverter_common/float16.py#L82

        # Send onnxmltools warnings to stdout (and therefore the log file)
        # so that they don't fill up the command line
        default_warnings = warnings.showwarning
        warnings.showwarning = _warn_to_stdout

        # Legalize ops are ops that have been or are currently in the block list
        # that we explicitly want removed
        legalize_ops = ["InstanceNormalization", "Resize", "Max"]
        op_block_list = onnxmltools.utils.float16_converter.DEFAULT_OP_BLOCK_LIST.copy()
        for op in legalize_ops:
            # Check to see that they are not in the block list before we remove them
            # Neccesary because the block list may be updated, and not in the state we expect
            if op in op_block_list:
                op_block_list.remove(op)

        # Infer shapes before converting to FP16 to enable models with >2GB
        onnx.shape_inference.infer_shapes_path(input_onnx)

        fp32_model = onnx.load_model(input_onnx)
        fp16_model = onnxmltools.utils.float16_converter.convert_float_to_float16(
            fp32_model, op_block_list=op_block_list, disable_shape_infer=True
        )

        # Load inputs and convert to fp16
        inputs_file = state.original_inputs_file
        if os.path.isfile(inputs_file):
            inputs = np.load(inputs_file, allow_pickle=True)
            inputs_converted = tensor_helpers.save_inputs(
                inputs, inputs_file, downcast=True
            )
        else:
            raise exp.StageError(
                "Attempted to convert inputs to FP16, however inputs file was not found."
            )

        # Overwrite expected dtypes
        _, state.expected_input_dtypes = build.get_shapes_and_dtypes(
            inputs_converted[0]
        )

        # Indicate that inputs must be downcasted during inference
        state.downcast_applied = True

        # Save FP16 model (use external data format if needed)
        output_path = converted_onnx_file(state)
        try:
            onnxmltools.utils.save_model(fp16_model, output_path)
        except ValueError:
            onnx.save_model(fp16_model, output_path, save_as_external_data=True)

        # Restore default warnings behavior
        warnings.showwarning = default_warnings

        # Check that the converted model is still valid
        success_msg = "\tSuccess converting ONNX model to fp16"
        fail_msg = "\tFailed converting ONNX model to fp16"

        if check_model(output_path, success_msg, fail_msg):
            state.intermediate_results = [output_path]

            stats = fs.Stats(
                state.cache_dir, state.config.build_name, state.evaluation_id
            )
            stats.save_model_eval_stat(
                fs.Keys.ONNX_FILE,
                output_path,
            )
        else:
            msg = f"""
            Attempted to use onnxmltools, a third party library, to convert your
            model to the float16 datatype, however this operation was not successful.
            More information may be available in the log file at **{self.logfile_path}**
            """
            raise exp.StageError(msg)

        return state



import argparse
from dataclasses import dataclass
from typing import TypeVar, List, Union


@dataclass
class Arg:
    """
    Class for tracking command-line arg configurations
    """

    name: str
    default: int
    type: TypeVar
    action: str = None

    def add_arg(self, parser_obj):
        if self.type == bool:
            parser_obj.add_argument(
                "--" + self.name, default=self.default, action=self.action
            )
        else:
            parser_obj.add_argument(
                "--" + self.name, default=self.default, type=self.type, nargs="?"
            )


def parse(valid_args: List[str]) -> List[Union[int, float]]:
    """
    The parse() function receives a list of strings with the desired arguments and
    returns a list containing the values of those arguments.

    The parse function uses argparse, allowing users to parse arguments through the CLI
    Using any argument that is not part of the list of valid args will result in a ValueError.
    """
    # List valid args
    tkml_args = {
        # General args
        # Batch size for the input to the model that will be used for benchmarking
        "batch_size": Arg("batch_size", default=1, type=int),
        # Maximum sequence length for the model's input;
        # also the input sequence length that will be used for benchmarking
        "max_seq_length": Arg("max_seq_length", default=128, type=int),
        # Maximum sequence length for the model's audio input;
        # also the input sequence length that will be used for benchmarking
        "max_audio_seq_length": Arg("max_audio_seq_length", default=25600, type=int),
        # Height of the input image that will be used for benchmarking
        "height": Arg("height", default=224, type=int),
        # Number of channels in the input image that will be used for benchmarking
        "num_channels": Arg("num_channels", default=3, type=int),
        # Width of the input image that will be used for benchmarking
        "width": Arg("width", default=224, type=int),
        # Args for Graph Neural Networks
        # k is Chebyshev filter is a polynomial-based filter that provides
        # a fast and localized spectral response on graph-structured data
        # example: used in chebconv
        "k": Arg("k", default=2, type=int),
        # alpha represents teleport probability used in the propagation step,
        # determining the amount of information considered from distant nodes in the graph
        "alpha": Arg("alpha", default=2.2, type=float),
        # Out channels represent the number of output features or
        # dimensions generated by the graph neural network layers.
        "out_channels": Arg("out_channels", default=1500, type=int),
        # Num_layers represent the number of hidden layers or depth neural
        # network, affecting its expressiveness and complexity.
        "num_layers": Arg("num_layers", default=8, type=int),
        # In_channels represent the number of input features per node in the
        # graph, defining the input dimensionality for the network.
        "in_channels": Arg("in_channels", default=1433, type=int),
        # Pretrained indicates whether pretrained weights should be used on the model
        "pretrained": Arg("pretrained", default=False, type=bool, action="store_true"),
    }

    # Create parser that accepts only the args received as part of valid_args
    parser = argparse.ArgumentParser()
    for arg_name in valid_args:
        if arg_name not in tkml_args.keys():
            raise KeyError(
                f"{arg_name} is not a valid arg. Valid args are {str(tkml_args.keys())}."
            )
        tkml_args[arg_name].add_arg(parser)

    # Parse arg and return args as a list
    args = vars(parser.parse_args())
    parsed_args = [args[arg] for arg in valid_args]
    return parsed_args



from setuptools import setup

with open("src/turnkeyml/version.py", encoding="utf-8") as fp:
    version = fp.read().split('"')[1]

setup(
    name="turnkeyml",
    version=version,
    description="TurnkeyML Tools and Models",
    author="Jeremy Fowers, Daniel Holanda, Ramakrishnan Sivakumar, Victoria Godsoe",
    author_email="jeremy.fowers@amd.com, daniel.holandanoronha@amd.com, krishna.sivakumar@amd.com, victoria.godsoe@amd.com",
    package_dir={"": "src", "turnkeyml_models": "models"},
    packages=[
        "turnkeyml",
        "turnkeyml.analyze",
        "turnkeyml.build",
        "turnkeyml.run",
        "turnkeyml.run.onnxrt",
        "turnkeyml.run.tensorrt",
        "turnkeyml.run.torchrt",
        "turnkeyml.cli",
        "turnkeyml.common",
        "turnkeyml_models",
        "turnkeyml_models.graph_convolutions",
        "turnkeyml_models.selftest",
        "turnkeyml_models.timm",
        "turnkeyml_models.torch_hub",
        "turnkeyml_models.torchvision",
        "turnkeyml_models.transformers",
    ],
    install_requires=[
        "invoke>=2.0.0",
        "onnx>=1.11.0",
        "onnxmltools==1.10.0",
        "hummingbird-ml==0.4.4",
        "scikit-learn==1.1.1",
        "xgboost==1.6.1",
        "lightgbm==3.3.5",
        "onnxruntime >=1.10.1, <1.16.0",
        "paramiko==2.11.0",
        "torch>=1.12.1",
        "protobuf>=3.17.3,<3.21",
        "pyyaml>=5.4",
        "typeguard>=2.3.13",
        "packaging>=20.9",
        "pandas>=1.5.3",
        "fasteners",
        "GitPython>=3.1.40",
    ],
    extras_require={
        "tensorflow": [
            "tensorflow-cpu==2.8.1",
            "tf2onnx>=1.12.0",
        ],
    },
    classifiers=[],
    entry_points={
        "console_scripts": [
            "turnkey=turnkeyml:turnkeycli",
        ]
    },
    python_requires=">=3.8, <3.11",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    package_data={
        "turnkeyml_models": ["requirements.txt", "readme.md"],
    },
)



import os
import logging
import sys
import pathlib
import copy
import traceback
import platform
import subprocess
import enum
from typing import Optional, Any, List, Dict, Union, Type
import dataclasses
import hashlib
import pkg_resources
import psutil
import yaml
import torch
import numpy as np
import sklearn.base
import turnkeyml.common.exceptions as exp
import turnkeyml.common.tf_helpers as tf_helpers
from turnkeyml.version import __version__ as turnkey_version


UnionValidModelInstanceTypes = Union[
    None,
    str,
    torch.nn.Module,
    torch.jit.ScriptModule,
    "tf.keras.Model",
    sklearn.base.BaseEstimator,
]

if os.environ.get("TURNKEY_ONNX_OPSET"):
    DEFAULT_ONNX_OPSET = int(os.environ.get("TURNKEY_ONNX_OPSET"))
else:
    DEFAULT_ONNX_OPSET = 14

MINIMUM_ONNX_OPSET = 11

DEFAULT_REBUILD_POLICY = "if_needed"
REBUILD_OPTIONS = ["if_needed", "always", "never"]


class ModelType(enum.Enum):
    PYTORCH = "pytorch"
    PYTORCH_COMPILED = "pytorch_compiled"
    KERAS = "keras"
    ONNX_FILE = "onnx_file"
    HUMMINGBIRD = "hummingbird"
    UNKNOWN = "unknown"


# Indicates that the build should take take any specific device into account
DEFAULT_DEVICE = "default"


def load_yaml(file_path):
    with open(file_path, "r", encoding="utf8") as stream:
        try:
            return yaml.load(stream, Loader=yaml.FullLoader)
        except yaml.YAMLError as e:
            raise exp.IOError(
                f"Failed while trying to open {file_path}."
                f"The exception that triggered this was:\n{e}"
            )


def output_dir(cache_dir, build_name):
    path = os.path.join(cache_dir, build_name)
    return path


def state_file(cache_dir, build_name):
    state_file_name = f"{build_name}_state.yaml"
    path = os.path.join(output_dir(cache_dir, build_name), state_file_name)
    return path


def hash_model(model, model_type: ModelType, hash_params: bool = True):
    # If the model is a path to a file, hash the file
    if model_type == ModelType.ONNX_FILE:
        # TODO: Implement a way of hashing the models but not the parameters
        # of ONNX inputs.
        if not hash_params:
            msg = "hash_params must be True for model_type ONNX_FILE"
            raise ValueError(msg)
        if os.path.isfile(model):
            with open(model, "rb") as f:
                file_content = f.read()
            return hashlib.sha256(file_content).hexdigest()
        else:
            raise ValueError(
                "hash_model received str model that doesn't correspond to a file"
            )

    elif model_type in [ModelType.PYTORCH, ModelType.PYTORCH_COMPILED]:
        # Convert model parameters and topology to string
        hashable_params = {}
        for name, param in model.named_parameters():
            hashable_params[name] = param.data
        if hash_params:
            hashable_model = (str(model) + str(hashable_params)).encode()
        else:
            hashable_model = str(model).encode()

        # Return hash of topology and parameters
        return hashlib.sha256(hashable_model).hexdigest()

    elif model_type == ModelType.KERAS:
        # Convert model parameters and topology to string
        summary_list = []  # type: List[str]

        # pylint: disable=unnecessary-lambda
        model.summary(print_fn=lambda x: summary_list.append(x))

        summary_str = " ".join(summary_list)
        hashable_params = {}
        for layer in model.layers:
            hashable_params[layer.name] = layer.weights
        if hash_params:
            hashable_model = (summary_str + str(hashable_params)).encode()
        else:
            hashable_model = summary_str.encode()

        # Return hash of topology and parameters
        return hashlib.sha256(hashable_model).hexdigest()

    elif model_type == ModelType.HUMMINGBIRD:
        import pickle

        return hashlib.sha256(pickle.dumps(model)).hexdigest()

    else:
        msg = f"""
        model_type "{model_type}" unsupported by this hash_model function
        """
        raise ValueError(msg)


class FunctionStatus(enum.Enum):
    """
    Status values that are assigned to stages, builds, benchmarks, and other
    functionality to help the user understand whether that function completed
    successfully or not.
    """

    # SUCCESSFUL means the stage/build/benchmark completed successfully.
    SUCCESSFUL = "successful"

    # ERROR means the stage/build/benchmark failed and threw some error that
    # was caught by turnkey. You should proceed by looking at the build
    # logs to see what happened.

    ERROR = "error"

    # KILLED means the build/benchmark failed because the system killed it. This can
    # happen because of an out-of-memory (OOM), timeout, system shutdown, etc.
    # You should proceed by re-running the build and keeping an eye on it to observe
    # why it is being killed (e.g., watch the RAM utilization to diagnose an OOM).
    KILLED = "killed"

    # The NOT_STARTED status is applied to all stages/builds/benchmarks at startup.
    # It will be replaced by one of the other status values if the stage/build/benchmark
    # has a chance to start running.
    # A value of NOT_STARTED in the report CSV indicates that the stage/build/benchmark
    # never had a chance to start because turnkey exited before that functionality had
    # a chance to start running.
    NOT_STARTED = "not_started"

    # INCOMPLETE indicates that a stage/build/benchmark started running and did not complete.
    # Each stage, build, and benchmark are marked as INCOMPLETE when they start running.
    # If you open the turnkey_stats.yaml file while the stage/build/benchmark
    # is still running, the status will show as INCOMPLETE. If the stage/build/benchmark
    # is killed without the chance to do any stats cleanup, the status will continue to
    # show as INCOMPLETE in turnkey_stats.yaml.
    # When the report CSV is created, any instance of an INCOMPLETE stage/build/benchmark
    # status will be replaced by KILLED.
    INCOMPLETE = "incomplete"


# Create a unique ID from this run by hashing pid + process start time
def unique_id():
    pid = os.getpid()
    p = psutil.Process(pid)
    start_time = p.create_time()
    return hashlib.sha256(f"{pid}{start_time}".encode()).hexdigest()


def get_shapes_and_dtypes(inputs: dict):
    """
    Return the shape and data type of each value in the inputs dict
    """
    shapes = {}
    dtypes = {}
    for key in sorted(inputs):
        value = inputs[key]
        if isinstance(
            value,
            (list, tuple),
        ):
            for v, i in zip(value, range(len(value))):
                subkey = f"{key}[{i}]"
                shapes[subkey] = np.array(v).shape
                dtypes[subkey] = np.array(v).dtype.name
        elif torch.is_tensor(value):
            shapes[key] = np.array(value.detach()).shape
            dtypes[key] = np.array(value.detach()).dtype.name
        elif tf_helpers.is_keras_tensor(value):
            shapes[key] = np.array(value).shape
            dtypes[key] = np.array(value).dtype.name
        elif isinstance(value, np.ndarray):
            shapes[key] = value.shape
            dtypes[key] = value.dtype.name
        elif isinstance(value, (bool, int, float)):
            shapes[key] = (1,)
            dtypes[key] = type(value).__name__
        elif value is None:
            pass
        else:
            raise exp.Error(
                "One of the provided inputs contains the unsupported "
                f' type {type(value)} at key "{key}".'
            )

    return shapes, dtypes


@dataclasses.dataclass(frozen=True)
class Config:
    """
    User-provided build configuration. Instances of Config should not be modified
    once they have been instantiated (frozen=True enforces this).

    Note: modifying this struct can create a breaking change that
    requires users to rebuild their models. Increment the minor
    version number of the turnkey package if you do make a build-
    breaking change.
    """

    build_name: str
    auto_name: bool
    sequence: List[str]
    onnx_opset: int
    device: Optional[str]


@dataclasses.dataclass
class State:
    # User-provided args that influence the generated model
    config: Config

    # User-provided args that do not influence the generated model
    monitor: bool = False
    rebuild: str = ""
    cache_dir: str = ""
    evaluation_id: str = ""

    # User-provided args that will not be saved as part of state.yaml
    model: UnionValidModelInstanceTypes = None
    inputs: Optional[Dict[str, Any]] = None

    # Member variable that helps the code know if State has called
    # __post_init__ yet
    save_when_setting_attribute: bool = False

    # All of the following are critical aspects of the build,
    # including properties of the tool and choices made
    # while building the model, which determine the outcome of the build.
    # NOTE: adding or changing a member name in this struct can create
    # a breaking change that requires users to rebuild their models.
    # Increment the minor version number of the turnkey package if you
    # do make a build-breaking change.

    turnkey_version: str = turnkey_version
    model_type: ModelType = ModelType.UNKNOWN
    uid: Optional[int] = None
    model_hash: Optional[int] = None
    build_status: FunctionStatus = FunctionStatus.NOT_STARTED
    expected_input_shapes: Optional[Dict[str, list]] = None
    expected_input_dtypes: Optional[Dict[str, list]] = None
    expected_output_names: Optional[List] = None

    # Whether or not inputs must be downcasted during inference
    downcast_applied: bool = False

    # The results of the most recent stage that was executed
    current_build_stage: str = None
    intermediate_results: Any = None

    # Results of a successful build
    results: Any = None

    def __post_init__(self):
        if self.uid is None:
            self.uid = unique_id()
        if self.inputs is not None:
            (
                self.expected_input_shapes,
                self.expected_input_dtypes,
            ) = get_shapes_and_dtypes(self.inputs)
        if self.model is not None and self.model_type != ModelType.UNKNOWN:
            self.model_hash = hash_model(self.model, self.model_type)

        self.save_when_setting_attribute = True

    def __setattr__(self, name, val):
        super().__setattr__(name, val)

        # Always automatically save the state.yaml whenever State is modified
        # But don't bother saving until after __post_init__ is done (indicated
        # by the save_when_setting_attribute flag)
        # Note: This only works when elements of the state are set directly.
        if self.save_when_setting_attribute and name != "save_when_setting_attribute":
            self.save()

    @property
    def original_inputs_file(self):
        return os.path.join(
            output_dir(self.cache_dir, self.config.build_name), "inputs.npy"
        )

    def prepare_file_system(self):
        # Create output folder if it doesn't exist
        os.makedirs(output_dir(self.cache_dir, self.config.build_name), exist_ok=True)

    def prepare_state_dict(self) -> Dict:
        state_dict = {
            key: value
            for key, value in vars(self).items()
            if not key == "inputs"
            and not key == "model"
            and not key == "save_when_setting_attribute"
        }

        # Special case for saving objects
        state_dict["config"] = copy.deepcopy(vars(self.config))

        state_dict["model_type"] = self.model_type.value
        state_dict["build_status"] = self.build_status.value

        return state_dict

    def save_yaml(self, state_dict: Dict):
        with open(
            state_file(self.cache_dir, self.config.build_name), "w", encoding="utf8"
        ) as outfile:
            yaml.dump(state_dict, outfile)

    def save(self):
        self.prepare_file_system()

        state_dict = self.prepare_state_dict()

        self.save_yaml(state_dict)


def load_state(
    cache_dir=None,
    build_name=None,
    state_path=None,
    state_type: Type = State,
) -> State:
    if state_path is not None:
        file_path = state_path
    elif build_name is not None and cache_dir is not None:
        file_path = state_file(cache_dir, build_name)
    else:
        raise ValueError(
            "This function requires either build_name and cache_dir to be set, "
            "or state_path to be set, not both or neither"
        )

    state_dict = load_yaml(file_path)

    # Get the type of Config and Info in case they have been overloaded
    field_types = {field.name: field.type for field in dataclasses.fields(state_type)}
    config_type = field_types["config"]

    try:
        # Special case for loading enums
        state_dict["model_type"] = ModelType(state_dict["model_type"])
        state_dict["build_status"] = FunctionStatus(state_dict["build_status"])
        state_dict["config"] = config_type(**state_dict["config"])

        state = state_type(**state_dict)

    except (KeyError, TypeError) as e:
        if state_path is not None:
            path_suggestion = pathlib.Path(state_path).parent
        else:
            path_suggestion = output_dir(cache_dir, build_name)
        msg = f"""
        The cached build of this model was built with an
        incompatible older version of the tool.

        Suggested solution: delete the build with
        rm -rf {path_suggestion}

        The underlying code raised this exception:
        {e}
        """
        raise exp.StateError(msg)

    return state


class Logger:
    """
    Redirects stdout to to file (and console if needed)
    """

    def __init__(
        self,
        initial_message: str,
        log_path: str = None,
    ):
        self.debug = os.environ.get("TURNKEY_BUILD_DEBUG") == "True"
        self.terminal = sys.stdout
        self.terminal_err = sys.stderr
        self.log_path = log_path

        # Create the empty logfile
        with open(log_path, "w", encoding="utf-8") as f:
            f.write(f"{initial_message}\n")

        # Disable any existing loggers so that we can capture all
        # outputs to a logfile
        self.root_logger = logging.getLogger()
        self.handlers = [handler for handler in self.root_logger.handlers]
        for handler in self.handlers:
            self.root_logger.removeHandler(handler)

        # Send any logger outputs to the logfile
        if not self.debug:
            self.file_handler = logging.FileHandler(filename=log_path)
            self.file_handler.setLevel(logging.INFO)
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
            self.file_handler.setFormatter(formatter)
            self.root_logger.addHandler(self.file_handler)

    def __enter__(self):
        sys.stdout = self
        sys.stderr = self

    def __exit__(self, _exc_type, _exc_value, _exc_tb):
        # Ensure we also capture the traceback as part of the logger when exceptions happen
        if _exc_type:
            traceback.print_exception(_exc_type, _exc_value, _exc_tb)

        # Stop redirecting stdout/stderr
        sys.stdout = self.terminal
        sys.stderr = self.terminal_err

        # Remove the logfile logging handler
        if not self.debug:
            self.file_handler.close()
            self.root_logger.removeHandler(self.file_handler)

            # Restore any pre-existing loggers
            for handler in self.handlers:
                self.root_logger.addHandler(handler)

    def write(self, message):
        if self.log_path is not None:
            with open(self.log_path, "a", encoding="utf-8") as f:
                f.write(message)
        if self.debug or self.log_path is None:
            self.terminal.write(message)
            self.terminal.flush()
            self.terminal_err.write(message)
            self.terminal_err.flush()

    def flush(self):
        # needed for python 3 compatibility.
        pass


def get_system_info():
    os_type = platform.system()
    info_dict = {}

    # Get OS Version
    try:
        info_dict["OS Version"] = platform.platform()
    except Exception as e:  # pylint: disable=broad-except
        info_dict["Error OS Version"] = str(e)

    if os_type == "Windows":
        # Get Processor Information
        try:
            proc_info = (
                subprocess.check_output("wmic cpu get name", shell=True)
                .decode()
                .split("\n")[1]
                .strip()
            )
            info_dict["Processor"] = proc_info
        except Exception as e:  # pylint: disable=broad-except
            info_dict["Error Processor"] = str(e)

        # Get OEM System Information
        try:
            oem_info = (
                subprocess.check_output("wmic computersystem get model", shell=True)
                .decode()
                .split("\n")[1]
                .strip()
            )
            info_dict["OEM System"] = oem_info
        except Exception as e:  # pylint: disable=broad-except
            info_dict["Error OEM System"] = str(e)

        # Get Physical Memory in GB
        try:
            mem_info_bytes = (
                subprocess.check_output(
                    "wmic computersystem get TotalPhysicalMemory", shell=True
                )
                .decode()
                .split("\n")[1]
                .strip()
            )
            mem_info_gb = round(int(mem_info_bytes) / (1024**3), 2)
            info_dict["Physical Memory"] = f"{mem_info_gb} GB"
        except Exception as e:  # pylint: disable=broad-except
            info_dict["Error Physical Memory"] = str(e)

    elif os_type == "Linux":
        # WSL has to be handled differently compared to native Linux
        if "microsoft" in str(platform.release()):
            try:
                oem_info = (
                    subprocess.check_output(
                        'powershell.exe -Command "wmic computersystem get model"',
                        shell=True,
                    )
                    .decode()
                    .strip()
                )
                oem_info = (
                    oem_info.replace("\r", "")
                    .replace("\n", "")
                    .split("Model")[-1]
                    .strip()
                )
                info_dict["OEM System"] = oem_info
            except Exception as e:  # pylint: disable=broad-except
                info_dict["Error OEM System (WSL)"] = str(e)

        else:
            # Get OEM System Information
            try:
                oem_info = (
                    subprocess.check_output(
                        "sudo dmidecode -s system-product-name",
                        shell=True,
                    )
                    .decode()
                    .strip()
                    .replace("\n", " ")
                )
                info_dict["OEM System"] = oem_info
            except Exception as e:  # pylint: disable=broad-except
                info_dict["Error OEM System"] = str(e)

        # Get CPU Information
        try:
            cpu_info = subprocess.check_output("lscpu", shell=True).decode()
            for line in cpu_info.split("\n"):
                if "Model name:" in line:
                    info_dict["Processor"] = line.split(":")[1].strip()
                    break
        except Exception as e:  # pylint: disable=broad-except
            info_dict["Error Processor"] = str(e)

        # Get Memory Information
        try:
            mem_info = (
                subprocess.check_output("free -m", shell=True)
                .decode()
                .split("\n")[1]
                .split()[1]
            )
            mem_info_gb = round(int(mem_info) / 1024, 2)
            info_dict["Memory Info"] = f"{mem_info_gb} GB"
        except Exception as e:  # pylint: disable=broad-except
            info_dict["Error Memory Info"] = str(e)

    else:
        info_dict["Error"] = "Unsupported OS"

    # Get Python Packages
    try:
        installed_packages = pkg_resources.working_set
        info_dict["Python Packages"] = [
            f"{i.key}=={i.version}"
            for i in installed_packages  # pylint: disable=not-an-iterable
        ]
    except Exception as e:  # pylint: disable=broad-except
        info_dict["Error Python Packages"] = str(e)

    return info_dict



# labels: test_group::turnkey name::dnaconv author::graph_convolutions task::Graph_Machine_Learning license::mit
"""
The dynamic neighborhood aggregation operator from the
`"Just Jump: Towards Dynamic Neighborhood Aggregation in Graph Neural Networks"
<https://arxiv.org/abs/1904.04849>`_ paper
"""

from turnkeyml.parser import parse
import torch

from torch_geometric.datasets import Planetoid
from torch_geometric.nn import DNAConv

dataset = Planetoid(root=".", name="Cora")
data = dataset[0]
edge_index_rows = 2

# Parsing command-line arguments
num_layers, in_channels = parse(["num_layers", "in_channels"])


model = DNAConv(in_channels)
# The input node features of shape :obj:`[num_nodes, num_layers, channels]`
x = torch.ones(data.num_nodes, num_layers, data.num_features, dtype=torch.float)
edge_index = torch.ones(edge_index_rows, data.num_nodes, dtype=torch.long)
inputs = {
    "x": x,
    "edge_index": edge_index,
}

# Call model
model(**inputs)



import os
import math
import platform
from typing import Dict, Union, List
from turnkeyml.common import printing
import turnkeyml.common.build as build
from turnkeyml.analyze.util import ModelInfo


def update(models_found: Dict[str, ModelInfo], build_name: str, cache_dir: str) -> None:
    """
    Prints all models and submodels found
    """
    if os.environ.get("TURNKEY_DEBUG") != "True":
        if platform.system() != "Windows":
            os.system("clear")
        else:
            os.system("cls")

    printing.logn(
        "\nModels discovered during profiling:\n",
        c=printing.Colors.BOLD,
    )
    recursive_print(models_found, build_name, cache_dir, None, None, [])


def recursive_print(
    models_found: Dict[str, ModelInfo],
    build_name: str,
    cache_dir: str,
    parent_model_hash: Union[str, None] = None,
    parent_invocation_hash: Union[str, None] = None,
    script_names_visited: List[str] = False,
) -> None:
    script_names_visited = []

    for model_hash in models_found.keys():
        model_visited = False
        model_info = models_found[model_hash]
        invocation_idx = 0
        for invocation_hash in model_info.unique_invocations.keys():
            unique_invocation = model_info.unique_invocations[invocation_hash]

            if (
                parent_model_hash == model_info.parent_hash
                and unique_invocation.executed > 0
                and (
                    model_info.unique_invocations[invocation_hash].parent_hash
                    == parent_invocation_hash
                )
            ):
                print_file_name = False
                if model_info.script_name not in script_names_visited:
                    script_names_visited.append(model_info.script_name)
                    if model_info.depth == 0:
                        print_file_name = True

                print_invocation(
                    model_info,
                    build_name,
                    cache_dir,
                    invocation_hash,
                    print_file_name,
                    invocation_idx=invocation_idx,
                    model_visited=model_visited,
                )
                model_visited = True
                invocation_idx += 1

                if print_file_name:
                    script_names_visited.append(model_info.script_name)

                recursive_print(
                    models_found,
                    build_name,
                    cache_dir,
                    parent_model_hash=model_hash,
                    parent_invocation_hash=invocation_hash,
                    script_names_visited=script_names_visited,
                )


def _pretty_print_key(key: str) -> str:
    result = key.split("_")
    result = [word.capitalize() for word in result]
    result = " ".join(result)
    return result


def parameters_to_size(parameters: int, byte_per_parameter: int = 4) -> str:
    size_bytes = parameters * byte_per_parameter
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def print_invocation(
    model_info: ModelInfo,
    build_name: str,
    cache_dir: str,
    invocation_hash: Union[str, None],
    print_file_name: bool = False,
    invocation_idx: int = 0,
    model_visited: bool = False,
) -> None:
    """
    Print information about a given model or submodel
    """
    unique_invocation = model_info.unique_invocations[invocation_hash]
    if model_info.model_type == build.ModelType.ONNX_FILE:
        extension = ".onnx"
        ident = "\t" * (2 * model_info.depth)
    else:
        extension = ".py"
        ident = "\t" * (2 * model_info.depth + 1)

    if print_file_name:
        print(f"{model_info.script_name}{extension}:")

    if unique_invocation.exec_time == 0 or model_info.build_model:
        exec_time = ""
    else:
        exec_time = f" - {unique_invocation.exec_time:.2f}s"

    # Print invocation about the model (only applies to scripts, not ONNX files)
    if model_info.model_type != build.ModelType.ONNX_FILE:
        if model_info.depth == 0 and len(model_info.unique_invocations) > 1:
            if not model_visited:
                printing.logn(f"{ident}{model_info.name}")
        else:
            printing.log(f"{ident}{model_info.name}")
            printing.logn(
                f" (executed {unique_invocation.executed}x{exec_time})",
                c=printing.Colors.OKGREEN,
            )

    if (model_info.depth == 0 and not model_visited) or (model_info.depth != 0):
        if model_info.depth == 0:
            if model_info.model_type == build.ModelType.PYTORCH:
                print(f"{ident}\tModel Type:\tPytorch (torch.nn.Module)")
            elif model_info.model_type == build.ModelType.KERAS:
                print(f"{ident}\tModel Type:\tKeras (tf.keras.Model)")
            elif model_info.model_type == build.ModelType.ONNX_FILE:
                print(f"{ident}\tModel Type:\tONNX File (.onnx)")

        # Display class of model and where it was found, if
        # the an input script (and not an input onnx file) was used
        if model_info.model_type != build.ModelType.ONNX_FILE:
            model_class = type(model_info.model)
            print(f"{ident}\tClass:\t\t{model_class.__name__} ({model_class})")
            if model_info.depth == 0:
                print(f"{ident}\tLocation:\t{model_info.file}, line {model_info.line}")

        # Display number of parameters and size
        parameters_size = parameters_to_size(model_info.params)
        print(
            f"{ident}\tParameters:\t{'{:,}'.format(model_info.params)} ({parameters_size})"
        )

    if model_info.depth == 0 and len(model_info.unique_invocations) > 1:
        printing.logn(
            f"\n{ident}\tWith input shape {invocation_idx+1} "
            f"(executed {unique_invocation.executed}x{exec_time})",
            c=printing.Colors.OKGREEN,
        )

    # Prepare input shape to be printed
    input_shape = dict(model_info.unique_invocations[invocation_hash].input_shapes)
    input_shape = {key: value for key, value in input_shape.items() if value != ()}
    input_shape = str(input_shape).replace("{", "").replace("}", "")

    print(f"{ident}\tInput Shape:\t{input_shape}")
    print(f"{ident}\tHash:\t\t" + invocation_hash)
    print(f"{ident}\tBuild dir:\t" + cache_dir + "/" + build_name)

    # Print turnkey results if turnkey was run
    if unique_invocation.performance:
        printing.log(f"{ident}\tStatus:\t\t")
        printing.logn(
            f"Successfully benchmarked on {unique_invocation.performance.device} "
            f"({unique_invocation.performance.runtime} "
            f"v{unique_invocation.performance.runtime_version}) ",
            c=unique_invocation.status_message_color,
        )
        printing.logn(
            f"{ident}\t\t\tMean Latency:\t{unique_invocation.performance.mean_latency:.3f}"
            f"\t{unique_invocation.performance.latency_units}"
        )
        printing.logn(
            f"{ident}\t\t\tThroughput:\t{unique_invocation.performance.throughput:.1f}"
            f"\t{unique_invocation.performance.throughput_units}"
        )

        if unique_invocation.stats_keys is not None:
            for key in unique_invocation.stats_keys:
                nice_key = _pretty_print_key(key)
                value = unique_invocation.stats.evaluation_stats[key]
                printing.logn(f"{ident}\t\t\t{nice_key}:\t{value}")
        print()
    else:
        if unique_invocation.is_target and model_info.build_model:
            printing.log(f"{ident}\tStatus:\t\t")
            printing.logn(
                f"{unique_invocation.status_message}",
                c=unique_invocation.status_message_color,
            )

            if unique_invocation.traceback is not None:
                if os.environ.get("TURNKEY_TRACEBACK") != "False":
                    for line in unique_invocation.traceback:
                        for subline in line.split("\n")[:-1]:
                            print(f"{ident}\t{subline}")

                else:
                    printing.logn(
                        f"{ident}\t\t\tTo see the full stack trace, "
                        "rerun with `export TURNKEY_TRACEBACK=True`.\n",
                        c=model_info.status_message_color,
                    )
            else:
                print()
        print()
